//
// Copyright (c) 2014, Facebook, Inc.
// All rights reserved.
//
// This source code is licensed under the University of Illinois/NCSA Open
// Source License found in the LICENSE file in the root directory of this
// source tree. An additional grant of patent rights can be found in the
// PATENTS file in the same directory.
//

#ifndef __DebugServer2_Types_h
#define __DebugServer2_Types_h

#include "DebugServer2/Base.h"
#include "DebugServer2/Constants.h"
#include "DebugServer2/CPUTypes.h"

#include <cstdint>
#include <map>
#include <stdlib.h>
#include <string>
#include <vector>
#if !defined(OS_WIN32)
#include <unistd.h>
#endif

namespace ds2 {

//
// Basic Types
//

#if defined(OS_WIN32)
typedef DWORD ProcessId;
typedef DWORD ThreadId;
typedef PSID UserId;
typedef PSID GroupId;
#else
typedef pid_t ProcessId;
typedef pid_t ThreadId;
typedef uid_t UserId;
typedef gid_t GroupId;
#endif

static const ProcessId kAllProcessId = static_cast<ProcessId>(-1);
static const ProcessId kAnyProcessId = static_cast<ProcessId>(0);
static const ThreadId kAllThreadId = static_cast<ThreadId>(-1);
static const ThreadId kAnyThreadId = static_cast<ThreadId>(0);

typedef std::vector<uint8_t> U8Vector;

//
// Process/Thread Id Tuple
//
struct ProcessThreadId {
  ProcessId pid;
  ThreadId tid;

  ProcessThreadId(ProcessId _pid = kAnyProcessId, ThreadId _tid = kAnyThreadId)
      : pid(_pid), tid(_tid) {}

  inline bool validPid() const {
    return pid != kAllProcessId && pid != kAnyProcessId;
  }

  inline bool validTid() const {
    return tid != kAllThreadId && tid != kAnyThreadId;
  }

  inline bool valid() const { return validPid() || validTid(); }

  inline bool any() const { return !validPid() && !validTid(); }

  inline void clear() {
    pid = kAnyProcessId;
    tid = kAnyThreadId;
  }
};

//
// STL Types
//

typedef std::vector<std::string> StringCollection;
typedef std::map<std::string, std::string> EnvironmentBlock;

//
// Repesents the stop information of a process.
//

struct StopInfo {
  enum Event {
    kEventNone,
    kEventStop,
    kEventExit,
    kEventKill,
  };

  enum Reason {
    kReasonNone,
    kReasonWatchpoint,
    kReasonRegisterWatchpoint,
    kReasonAddressWatchpoint,
    kReasonLibraryLoad,
    kReasonLibraryUnload,
    kReasonReplayLog,
    kReasonBreakpoint,
    kReasonTrace,
    kReasonSignalStop, // TODO better name
    kReasonException,
    kReasonTrap,
  };

  Event event;
  Reason reason;
  uint32_t core;
  int status;
#if !defined(OS_WIN32)
  int signal;
#endif

  StopInfo() { clear(); }

  inline void clear() {
    event = kEventNone;
    reason = kReasonNone;
    core = 0;
    status = 0;
#if !defined(OS_WIN32)
    signal = 0;
#endif
  }
};

//
// Represents an address
//

class Address {
private:
  uint64_t _value;
  bool _unset;

public:
  Address() : _value(0), _unset(true) {}

  Address(uint64_t address) : _value(address), _unset(false) {}

public:
  inline operator uint64_t() const { return _value; }
  inline Address &operator=(uint64_t address) {
    _unset = false, _value = address;
    return *this;
  }

public:
  inline bool valid() const { return !_unset; }

public:
  inline uint64_t value() const { return _value; }

public:
  inline void unset() { _unset = true, _value = 0; }
  inline void clear() { unset(); }
};

//
// Constant for invalid native CPU (sub)type.
//
static uint32_t const kInvalidCPUType = static_cast<uint32_t>(-2);

//
// Architecture Flags
//
enum {
  kArchFlagWatchpointExceptionsReceivedAfter = (0 << 0),
  kArchFlagWatchpointExceptionsReceivedBefore = (1 << 0)
};

//
// Describes the host information.
//
struct HostInfo {
  CPUType cpuType;
  CPUSubType cpuSubType;
  uint32_t nativeCPUType;
  uint32_t nativeCPUSubType;
  std::string hostName;
  std::string osType;
  std::string osVendor;
  std::string osBuild;
  std::string osKernel;
  std::string osVersion;
  Endian endian;
  size_t pointerSize;
  uint32_t archFlags;

  HostInfo() { clear(); }

  inline void clear() {
    cpuType = kCPUTypeAny;
    cpuSubType = kCPUSubTypeInvalid;
    nativeCPUType = kInvalidCPUType;
    nativeCPUSubType = kInvalidCPUType;
    endian = kEndianUnknown;
    pointerSize = 0;
    archFlags = 0;
  }
};

//
// Describe a process
//
struct ProcessInfo {
  typedef std::vector<ProcessInfo> Collection;

  ProcessId pid;
#if !defined(OS_WIN32)
  ProcessId parentPid;
#endif

  std::string name;

  UserId realUid;
  GroupId realGid;
#if !defined(OS_WIN32)
  UserId effectiveUid;
  GroupId effectiveGid;
#endif

  CPUType cpuType;
  CPUSubType cpuSubType;
  uint32_t nativeCPUType;
  uint32_t nativeCPUSubType;
  Endian endian;
  size_t pointerSize;
  uint32_t archFlags;

  std::string osType;
  std::string osVendor;

  ProcessInfo() { clear(); }

  inline void clear() {
    pid = kAnyProcessId;
#if !defined(OS_WIN32)
    parentPid = kAnyProcessId;
#endif

    name.clear();

#if defined(OS_WIN32)
    realUid = nullptr;
    realGid = nullptr;
#else
    realUid = 0;
    realGid = 0;
    effectiveUid = 0;
    effectiveGid = 0;
#endif

    cpuType = kCPUTypeAny;
    cpuSubType = kCPUSubTypeInvalid;
    nativeCPUType = kInvalidCPUType;
    nativeCPUSubType = kInvalidCPUType;
    endian = kEndianUnknown;
    pointerSize = 0;
    archFlags = 0;

    osType.clear();
    osVendor.clear();
  }

#if defined(OS_WIN32)
  ~ProcessInfo() {
    free(realUid);
    free(realGid);
  }
#endif
};

//
// Describes a memory region
//
struct MemoryRegionInfo {
  typedef std::vector<MemoryRegionInfo> Collection;

  Address start;
  uint64_t length;
  uint32_t protection;

  MemoryRegionInfo() { clear(); }

  inline void clear() {
    start.clear();
    length = 0;
    protection = 0;
  }
};

struct SharedLibraryInfo {
  bool main;
  std::string path;
  struct {
    uint64_t mapAddress;
    uint64_t baseAddress;
    uint64_t ldAddress;
  } svr4;
  std::vector<uint64_t> sections;
};
}

#endif // !__DebugServer2_Types_h
